#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2022 - 2023

import time
import uuid

from enum import Enum

from idds.common.utils import json_dumps


class EventBusState(Enum):
    New = 0
    Master = 1
    Slave = 2
    Unknown = 3


class EventType(Enum):
    Event = 0
    StateClaim = 1
    Demand = 2

    NewRequest = 10
    UpdateRequest = 11
    AbortRequest = 12
    ResumeRequest = 13
    ExpireRequest = 14

    NewTransform = 20
    UpdateTransform = 21
    AbortTransform = 22
    ResumeTransform = 23

    NewProcessing = 30
    UpdateProcessing = 31
    AbortProcessing = 32
    ResumeProcessing = 33
    SyncProcessing = 34
    TerminatedProcessing = 35
    TriggerProcessing = 36
    MsgTriggerProcessing = 37

    UpdateCommand = 40


class Event(object):
    def __init__(self, publisher_id, event_type=EventType.Event, content=None, counter=1):
        self._id = str(uuid.uuid4())
        self._publisher_id = publisher_id
        self._event_type = event_type
        self._timestamp = time.time()
        self._counter = counter
        self._content = content

    def to_json(self):
        ret = {'id': self._id, 'publisher_id': self._publisher_id,
               'event_type': (self._event_type.name, self._event_type.value),
               'timestamp': self._timestamp,
               'counter': self._counter,
               'content': self._content}
        return ret

    def __str__(self):
        return json_dumps(self.to_json())


class StateClaimEvent(Event):
    def __init__(self, publisher_id, event_bus_state, content=None, counter=1):
        super(StateClaimEvent, self).__init__(publisher_id, event_type=EventType.StateClaim, content=content, counter=counter)
        self._event_bus_state = event_bus_state

    def to_json(self):
        ret = super(StateClaimEvent, self).to_json()
        ret['event_bus_state'] = self._event_bus_state
        return ret


class DemandEvent(Event):
    def __init__(self, publisher_id, demand_type, content=None, counter=1):
        super(DemandEvent, self).__init__(publisher_id, event_type=EventType.Demand, content=content, counter=counter)
        self._demand_type = demand_type

    def to_json(self):
        ret = super(DemandEvent, self).to_json()
        ret['demand_type'] = self._demand_type
        return ret


class NewRequestEvent(Event):
    def __init__(self, publisher_id, request_id, content=None, counter=1):
        super(NewRequestEvent, self).__init__(publisher_id, event_type=EventType.NewRequest, content=content, counter=counter)
        self._request_id = request_id

    def to_json(self):
        ret = super(NewRequestEvent, self).to_json()
        ret['request_id'] = self._request_id
        return ret


class UpdateRequestEvent(Event):
    def __init__(self, publisher_id, request_id, content=None, counter=1):
        super(UpdateRequestEvent, self).__init__(publisher_id, event_type=EventType.UpdateRequest, content=content, counter=counter)
        self._request_id = request_id

    def to_json(self):
        ret = super(UpdateRequestEvent, self).to_json()
        ret['request_id'] = self._request_id
        return ret


class AbortRequestEvent(Event):
    def __init__(self, publisher_id, request_id, content=None, counter=1):
        super(AbortRequestEvent, self).__init__(publisher_id, event_type=EventType.AbortRequest, content=content, counter=counter)
        self._request_id = request_id

    def to_json(self):
        ret = super(AbortRequestEvent, self).to_json()
        ret['request_id'] = self._request_id
        return ret


class ResumeRequestEvent(Event):
    def __init__(self, publisher_id, request_id, content=None, counter=1):
        super(ResumeRequestEvent, self).__init__(publisher_id, event_type=EventType.ResumeRequest, content=content, counter=counter)
        self._request_id = request_id

    def to_json(self):
        ret = super(ResumeRequestEvent, self).to_json()
        ret['request_id'] = self._request_id
        return ret


class ExpireRequestEvent(Event):
    def __init__(self, publisher_id, request_id, content=None, counter=1):
        super(ExpireRequestEvent, self).__init__(publisher_id, event_type=EventType.ExpireRequest, content=content, counter=counter)
        self._request_id = request_id

    def to_json(self):
        ret = super(ExpireRequestEvent, self).to_json()
        ret['request_id'] = self._request_id
        return ret


class UpdateCommandEvent(Event):
    def __init__(self, publisher_id, command_id, content=None, counter=1):
        super(UpdateCommandEvent, self).__init__(publisher_id, event_type=EventType.UpdateCommand, content=content, counter=counter)
        self._command_id = command_id

    def to_json(self):
        ret = super(UpdateCommandEvent, self).to_json()
        ret['command_id'] = self._command_id
        return ret


class NewTransformEvent(Event):
    def __init__(self, publisher_id, transform_id, content=None, counter=1):
        super(NewTransformEvent, self).__init__(publisher_id, event_type=EventType.NewTransform, content=content, counter=counter)
        self._transform_id = transform_id

    def to_json(self):
        ret = super(NewTransformEvent, self).to_json()
        ret['transform_id'] = self._transform_id
        return ret


class UpdateTransformEvent(Event):
    def __init__(self, publisher_id, transform_id, content=None, counter=1):
        super(UpdateTransformEvent, self).__init__(publisher_id, event_type=EventType.UpdateTransform, content=content, counter=counter)
        self._transform_id = transform_id

    def to_json(self):
        ret = super(UpdateTransformEvent, self).to_json()
        ret['transform_id'] = self._transform_id
        return ret


class AbortTransformEvent(Event):
    def __init__(self, publisher_id, transform_id, content=None, counter=1):
        super(AbortTransformEvent, self).__init__(publisher_id, event_type=EventType.AbortTransform, content=content, counter=counter)
        self._transform_id = transform_id

    def to_json(self):
        ret = super(AbortTransformEvent, self).to_json()
        ret['transform_id'] = self._transform_id
        return ret


class ResumeTransformEvent(Event):
    def __init__(self, publisher_id, transform_id, content=None, counter=1):
        super(ResumeTransformEvent, self).__init__(publisher_id, event_type=EventType.ResumeTransform, content=content, counter=counter)
        self._transform_id = transform_id

    def to_json(self):
        ret = super(ResumeTransformEvent, self).to_json()
        ret['transform_id'] = self._transform_id
        return ret


class NewProcessingEvent(Event):
    def __init__(self, publisher_id, processing_id, content=None, counter=1):
        super(NewProcessingEvent, self).__init__(publisher_id, event_type=EventType.NewProcessing, content=content, counter=counter)
        self._processing_id = processing_id

    def to_json(self):
        ret = super(NewProcessingEvent, self).to_json()
        ret['processing_id'] = self._processing_id
        return ret


class UpdateProcessingEvent(Event):
    def __init__(self, publisher_id, processing_id, content=None, counter=1):
        super(UpdateProcessingEvent, self).__init__(publisher_id, event_type=EventType.UpdateProcessing, content=content, counter=counter)
        self._processing_id = processing_id

    def to_json(self):
        ret = super(UpdateProcessingEvent, self).to_json()
        ret['processing_id'] = self._processing_id
        return ret


class AbortProcessingEvent(Event):
    def __init__(self, publisher_id, processing_id, content=None, counter=1):
        super(AbortProcessingEvent, self).__init__(publisher_id, event_type=EventType.AbortProcessing, content=content, counter=counter)
        self._processing_id = processing_id

    def to_json(self):
        ret = super(AbortProcessingEvent, self).to_json()
        ret['processing_id'] = self._processing_id
        return ret


class ResumeProcessingEvent(Event):
    def __init__(self, publisher_id, processing_id, content=None, counter=1):
        super(ResumeProcessingEvent, self).__init__(publisher_id, event_type=EventType.ResumeProcessing, content=content, counter=counter)
        self._processing_id = processing_id

    def to_json(self):
        ret = super(ResumeProcessingEvent, self).to_json()
        ret['processing_id'] = self._processing_id
        return ret


class SyncProcessingEvent(Event):
    def __init__(self, publisher_id, processing_id, content=None, counter=1):
        super(SyncProcessingEvent, self).__init__(publisher_id, event_type=EventType.SyncProcessing, content=content, counter=counter)
        self._processing_id = processing_id

    def to_json(self):
        ret = super(SyncProcessingEvent, self).to_json()
        ret['processing_id'] = self._processing_id
        return ret


class TerminatedProcessingEvent(Event):
    def __init__(self, publisher_id, processing_id, content=None, counter=1):
        super(TerminatedProcessingEvent, self).__init__(publisher_id, event_type=EventType.TerminatedProcessing, content=content, counter=counter)
        self._processing_id = processing_id

    def to_json(self):
        ret = super(TerminatedProcessingEvent, self).to_json()
        ret['processing_id'] = self._processing_id
        return ret


class TriggerProcessingEvent(Event):
    def __init__(self, publisher_id, processing_id, content=None, counter=1):
        super(TriggerProcessingEvent, self).__init__(publisher_id, event_type=EventType.TriggerProcessing, content=content, counter=counter)
        self._processing_id = processing_id

    def to_json(self):
        ret = super(TriggerProcessingEvent, self).to_json()
        ret['processing_id'] = self._processing_id
        return ret


class MsgTriggerProcessingEvent(Event):
    def __init__(self, publisher_id, processing_id, content=None, counter=1):
        super(MsgTriggerProcessingEvent, self).__init__(publisher_id, event_type=EventType.MsgTriggerProcessing, content=content, counter=counter)
        self._processing_id = processing_id

    def to_json(self):
        ret = super(MsgTriggerProcessingEvent, self).to_json()
        ret['processing_id'] = self._processing_id
        return ret



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020


"""
Test cacher.
"""

import os

from idds.client.client import Client
from idds.common.utils import get_rest_host


full_name = os.path.abspath(__file__)

host = get_rest_host()

client = Client(host=host)
client.upload(full_name)
client.download(os.path.join('/tmp', os.path.basename(full_name)))

full_name = '/bin/hostname'
client.upload(full_name)
client.download(os.path.join('/tmp', os.path.basename(full_name)))



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2021


from traceback import format_exc

from flask import Blueprint

from idds.common import exceptions
from idds.common.constants import (HTTP_STATUS_CODE, MessageType, MessageStatus,
                                   MessageSource, MessageDestination,
                                   CommandType, RequestStatus)
from idds.common.utils import json_loads
from idds.core.commands import add_command
from idds.core.requests import get_requests
from idds.core.messages import add_message, retrieve_messages
from idds.rest.v1.controller import IDDSController


class Message(IDDSController):
    """ Get message """

    def get(self, request_id, workload_id):
        """ Get messages with given id.
        HTTP Success:
            200 OK
        HTTP Error:
            404 Not Found
            500 InternalError
        :returns: dictionary of an request.
        """

        try:
            if request_id == 'null':
                request_id = None
            if workload_id == 'null':
                workload_id = None

            if request_id is None:
                raise Exception("request_id should not be None")
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.BadRequest, exc_cls=exceptions.BadRequest.__name__, exc_msg=str(error))

        try:
            username = self.get_username()
            reqs = get_requests(request_id=request_id, workload_id=workload_id, with_request=True)
            for req in reqs:
                if req['username'] and req['username'] != username:
                    raise exceptions.AuthenticationNoPermission("User %s has no permission to update request %s" % (username, req['request_id']))
        except exceptions.AuthenticationNoPermission as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        try:
            msgs = retrieve_messages(request_id=request_id, workload_id=workload_id)
            rets = []
            for msg in msgs:
                rets.append(msg['msg_content'])
        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data=rets)

    def post(self, request_id, workload_id):
        """ Create Request.
        HTTP Success:
            200 OK
        HTTP Error:
            400 Bad request
            500 Internal Error
        """
        try:
            if request_id == 'null':
                request_id = None
            if workload_id == 'null':
                workload_id = None
            if request_id is None:
                raise Exception("request_id should not be None")
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.BadRequest, exc_cls=exceptions.BadRequest.__name__, exc_msg=str(error))

        try:
            username = self.get_username()
            reqs = get_requests(request_id=request_id, workload_id=workload_id, with_request=True)
            for req in reqs:
                if req['username'] and req['username'] != username:
                    raise exceptions.AuthenticationNoPermission("User %s has no permission to update request %s" % (username, req['request_id']))
        except exceptions.AuthenticationNoPermission as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        try:
            msg = self.get_request().data and json_loads(self.get_request().data)
            # command = msg['command']
            # parameters = msg['parameters']
            if 'command' in msg and msg['command'] in ['update_request', 'update_processing']:
                status = msg['parameters']['status']
                if status in [RequestStatus.ToCancel, RequestStatus.ToSuspend]:
                    add_command(request_id=request_id, cmd_type=CommandType.AbortRequest,
                                cmd_content=None)
                elif status in [RequestStatus.ToResume]:
                    add_command(request_id=request_id, cmd_type=CommandType.ResumeRequest,
                                cmd_content=None)
            else:
                add_message(msg_type=MessageType.IDDSCommunication,
                            status=MessageStatus.New,
                            destination=MessageDestination.Clerk,
                            source=MessageSource.Rest,
                            request_id=request_id,
                            workload_id=workload_id,
                            transform_id=None,
                            num_contents=1,
                            msg_content=msg)

        except exceptions.DuplicatedObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.Conflict, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data={'request_id': request_id})

    def post_test(self):
        import pprint
        pprint.pprint(self.get_request())
        pprint.pprint(self.get_request().endpoint)
        pprint.pprint(self.get_request().url_rule)


"""----------------------
   Web service url maps
----------------------"""


def get_blueprint():
    bp = Blueprint('message', __name__)

    view = Message.as_view('message')
    bp.add_url_rule('/message/<request_id>/<workload_id>', view_func=view, methods=['get', 'post'])
    return bp



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Sergey Padolski, <spadolski@bnl.gov>, 2021
# - Wen Guan, <wen.guan@cern.ch>, 2021


"""
Test client.
"""

import string
import random
import time

# import traceback

# from rucio.client.client import Client as Rucio_Client
# from rucio.common.exception import CannotAuthenticate

# from idds.client.client import Client
from idds.client.clientmanager import ClientManager
# from idds.common.constants import RequestType, RequestStatus
from idds.common.utils import get_rest_host
# from idds.tests.common import get_example_real_tape_stagein_request
# from idds.tests.common import get_example_prodsys2_tape_stagein_request

# from idds.workflowv2.work import Work, Parameter, WorkStatus
# from idds.workflowv2.workflow import Condition, Workflow
from idds.workflowv2.workflow import Workflow
# from idds.atlas.workflowv2.atlasstageinwork import ATLASStageinWork
from idds.doma.workflowv2.domapandawork import DomaPanDAWork

# task_cloud = 'LSST'
task_cloud = 'US'

task_queue = 'DOMA_LSST_GOOGLE_TEST'
# task_queue = 'DOMA_LSST_GOOGLE_MERGE'
# task_queue = 'SLAC_TEST'
# task_queue = 'DOMA_LSST_SLAC_TEST'
task_queue = 'SLAC_Rubin'

# task_cloud = 'EU'
# task_queue = 'CC-IN2P3_TEST'

# task_cloud = 'EU'
# task_queue = 'LANCS_TEST'


def randStr(chars=string.ascii_lowercase + string.digits, N=10):
    return ''.join(random.choice(chars) for _ in range(N))


class PanDATask(object):
    name = None
    step = None
    dependencies = []


def setup_workflow():

    taskN1 = PanDATask()
    taskN1.step = "step1"
    taskN1.name = taskN1.step + "_" + randStr()
    taskN1.dependencies = [
        {"name": "00000" + str(k),
         "dependencies": [],
         "submitted": False} for k in range(6)
    ]

    taskN2 = PanDATask()
    taskN2.step = "step2"
    taskN2.name = taskN2.step + "_" + randStr()
    taskN2.dependencies = [
        {
            "name": "000010",
            "dependencies": [{"task": taskN1.name, "inputname": "000001", "available": False},
                             {"task": taskN1.name, "inputname": "000002", "available": False}],
            "submitted": False
        },
        {
            "name": "000011",
            "dependencies": [{"task": taskN1.name, "inputname": "000001", "available": False},
                             {"task": taskN1.name, "inputname": "000002", "available": False}],
            "submitted": False
        },
        {
            "name": "000012",
            "dependencies": [{"task": taskN1.name, "inputname": "000001", "available": False},
                             {"task": taskN1.name, "inputname": "000002", "available": False}],
            "submitted": False
        }
    ]

    taskN3 = PanDATask()
    taskN3.step = "step3"
    taskN3.name = taskN3.step + "_" + randStr()
    taskN3.dependencies = [
        {
            "name": "000020",
            "dependencies": [],
            "submitted": False
        },
        {
            "name": "000021",
            "dependencies": [{"task": taskN2.name, "inputname": "000010", "available": False},
                             {"task": taskN2.name, "inputname": "000011", "available": False}],
            "submitted": False
        },
        {
            "name": "000022",
            "dependencies": [{"task": taskN2.name, "inputname": "000011", "available": False},
                             {"task": taskN2.name, "inputname": "000012", "available": False}],
            "submitted": False
        },
        {
            "name": "000023",
            "dependencies": [],
            "submitted": False
        },
        {
            "name": "000024",
            "dependencies": [{"task": taskN3.name, "inputname": "000021", "available": False},
                             {"task": taskN3.name, "inputname": "000023", "available": False}],
            "submitted": False
        },
    ]

    work1 = DomaPanDAWork(executable='echo',
                          primary_input_collection={'scope': 'pseudo_dataset', 'name': 'pseudo_input_collection#1'},
                          output_collections=[{'scope': 'pseudo_dataset', 'name': 'pseudo_output_collection#1'}],
                          log_collections=[], dependency_map=taskN1.dependencies,
                          task_name=taskN1.name, task_queue=task_queue,
                          encode_command_line=True,
                          prodSourceLabel='managed',
                          task_log={"dataset": "PandaJob_#{pandaid}/",
                                    "destination": "local",
                                    "param_type": "log",
                                    "token": "local",
                                    "type": "template",
                                    "value": "log.tgz"},
                          task_cloud=task_cloud)
    work2 = DomaPanDAWork(executable='echo',
                          primary_input_collection={'scope': 'pseudo_dataset', 'name': 'pseudo_input_collection#2'},
                          output_collections=[{'scope': 'pseudo_dataset', 'name': 'pseudo_output_collection#2'}],
                          log_collections=[], dependency_map=taskN2.dependencies,
                          task_name=taskN2.name, task_queue=task_queue,
                          encode_command_line=True,
                          prodSourceLabel='managed',
                          task_log={"dataset": "PandaJob_#{pandaid}/",
                                    "destination": "local",
                                    "param_type": "log",
                                    "token": "local",
                                    "type": "template",
                                    "value": "log.tgz"},
                          task_cloud=task_cloud)
    work3 = DomaPanDAWork(executable='echo',
                          primary_input_collection={'scope': 'pseudo_dataset', 'name': 'pseudo_input_collection#3'},
                          output_collections=[{'scope': 'pseudo_dataset', 'name': 'pseudo_output_collection#3'}],
                          log_collections=[], dependency_map=taskN3.dependencies,
                          task_name=taskN3.name, task_queue=task_queue,
                          encode_command_line=True,
                          prodSourceLabel='managed',
                          task_log={"dataset": "PandaJob_#{pandaid}/",
                                    "destination": "local",
                                    "param_type": "log",
                                    "token": "local",
                                    "type": "template",
                                    "value": "log.tgz"},
                          task_cloud=task_cloud)

    pending_time = 12
    # pending_time = None
    workflow = Workflow(pending_time=pending_time)
    workflow.add_work(work1)
    workflow.add_work(work2)
    workflow.add_work(work3)
    workflow.name = 'test_workflow.idds.%s.test' % time.time()
    return workflow


if __name__ == '__main__':
    host = get_rest_host()
    workflow = setup_workflow()

    wm = ClientManager(host=host)
    # wm.set_original_user(user_name="wguandev")
    request_id = wm.submit(workflow, use_dataset_name=False)
    print(request_id)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2021 - 2022

import argparse
import sys
import os.path
base_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(base_path)
os.chdir(base_path)

from idds.orm.base.utils import destroy_database, destroy_everything      # noqa E402


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--destroy-everything", action="store_true", default=False, help='Destroy all tables+constraints')
    args = parser.parse_args()
    if args.destroy_everything:
        destroy_everything()
    else:
        destroy_database()



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019

import datetime
from traceback import format_exc

from flask import Blueprint

from idds.common import exceptions
from idds.common.constants import HTTP_STATUS_CODE
from idds.core.requests import get_requests
from idds.rest.v1.controller import IDDSController


class Monitor(IDDSController):
    """ Monitor """

    def get_month_list(self, start, end):
        mlist = []
        total_months = lambda dt: dt.month + 12 * dt.year
        for tot_m in range(total_months(start) - 1, total_months(end)):
            y, m = divmod(tot_m, 12)
            mlist.append(datetime.datetime(y, m + 1, 1).strftime("%Y-%m"))
        return mlist

    def get_requests(self, request_id, workload_id, with_request=False, with_transform=False, with_processing=False):

        if with_request:
            rets, ret_reqs = [], {}
            reqs = get_requests(request_id=request_id, workload_id=workload_id,
                                with_request=False, with_transform=True, with_processing=with_processing,
                                with_detail=False, with_metadata=False)
            for req in reqs:
                if req['request_id'] not in ret_reqs:
                    ret_reqs[req['request_id']] = {'request_id': req['request_id'],
                                                   'workload_id': req['workload_id'],
                                                   'status': req['status'].name if req['status'] else req['status'],
                                                   'created_at': req['created_at'],
                                                   'updated_at': req['updated_at'],
                                                   'transforms': {},
                                                   'input_total_files': 0,
                                                   'input_coll_bytes': 0,
                                                   'input_processed_files': 0,
                                                   'input_processing_files': 0,
                                                   'output_total_files': 0,
                                                   'output_coll_bytes': 0,
                                                   'output_processed_files': 0,
                                                   'output_processing_files': 0
                                                   }
                if req['transform_status']:
                    if req['transform_status'].name not in ret_reqs[req['request_id']]['transforms']:
                        ret_reqs[req['request_id']]['transforms'][req['transform_status'].name] = 0
                    ret_reqs[req['request_id']]['transforms'][req['transform_status'].name] += 1

                    if req['input_total_files']:
                        ret_reqs[req['request_id']]['input_total_files'] += req['input_total_files']
                    if req['input_coll_bytes']:
                        ret_reqs[req['request_id']]['input_coll_bytes'] += req['input_coll_bytes']
                    if req['input_processed_files']:
                        ret_reqs[req['request_id']]['input_processed_files'] += req['input_processed_files']
                    if req['input_processing_files']:
                        ret_reqs[req['request_id']]['input_processing_files'] += req['input_processing_files']

                    if req['output_total_files']:
                        ret_reqs[req['request_id']]['output_total_files'] += req['output_total_files']
                    if req['output_coll_bytes']:
                        ret_reqs[req['request_id']]['output_coll_bytes'] += req['output_coll_bytes']
                    if req['output_processed_files']:
                        ret_reqs[req['request_id']]['output_processed_files'] += req['output_processed_files']
                    if req['output_processing_files']:
                        ret_reqs[req['request_id']]['output_processing_files'] += req['output_processing_files']

            for req_id in ret_reqs:
                rets.append(ret_reqs[req_id])
            return rets
        elif with_transform:
            rets = []
            reqs = get_requests(request_id=request_id, workload_id=workload_id,
                                with_request=with_request, with_transform=with_transform, with_processing=with_processing,
                                with_detail=False, with_metadata=False)
            for req in reqs:
                ret = {'request_id': req['request_id'],
                       'transform_id': req['transform_id'],
                       'workload_id': req['workload_id'],
                       'transform_workload_id': req['transform_workload_id'],
                       'transform_type': req['transform_type'].name if req['transform_type'] else req['transform_type'],
                       'output_coll_scope': req['output_coll_scope'],
                       'output_coll_name': req['output_coll_name'],
                       'transform_status': req['transform_status'].name if req['transform_status'] else req['transform_status'],
                       'transform_created_at': req['transform_created_at'],
                       'transform_updated_at': req['transform_updated_at'],
                       'transform_finished_at': req['transform_finished_at'],
                       'input_total_files': req['input_total_files'] if req['input_total_files'] else 0,
                       'input_coll_bytes': req['input_coll_bytes'] if req['input_coll_bytes'] else 0,
                       'input_processed_files': req['input_processed_files'] if req['input_processed_files'] else 0,
                       'input_processing_files': req['input_processing_files'] if req['input_processing_files'] else 0,
                       'output_total_files': req['output_total_files'] if req['output_total_files'] else 0,
                       'output_coll_bytes': req['output_coll_bytes'] if req['output_coll_bytes'] else 0,
                       'output_processed_files': req['output_processed_files'] if req['output_processed_files'] else 0,
                       'output_processing_files': req['output_processing_files'] if req['output_processing_files'] else 0,
                       'errors': req['errors']
                       }
                rets.append(ret)
            return rets
        elif with_processing:
            rets = []
            reqs = get_requests(request_id=request_id, workload_id=workload_id,
                                with_request=with_request, with_transform=with_transform, with_processing=with_processing,
                                with_detail=False, with_metadata=False)
            for req in reqs:
                ret = {'request_id': req['request_id'],
                       'workload_id': req['processing_workload_id'],
                       'processing_id': req['processing_id'],
                       'processing_status': req['processing_status'].name if req['processing_status'] else req['processing_status'],
                       'processing_created_at': req['processing_created_at'],
                       'processing_updated_at': req['processing_updated_at'],
                       'processing_finished_at': req['processing_finished_at']
                       }
                rets.append(ret)
            return rets
        else:
            rets = []
            reqs = get_requests(request_id=request_id, workload_id=workload_id, with_detail=False, with_processing=False, with_metadata=False)
            for req in reqs:
                ret = {'request_id': req['request_id'],
                       'workload_id': req['workload_id'],
                       'status': req['status'].name if req['status'] else req['status'],
                       'created_at': req['created_at'],
                       'updated_at': req['updated_at']
                       }
                rets.append(ret)
            return rets

    def get(self, request_id, workload_id, with_request='false', with_transform='false', with_processing='false'):
        """ Get details about a specific Request with given id.
        HTTP Success:
            200 OK
        HTTP Error:
            404 Not Found
            500 InternalError
        :returns: dictionary of an request.
        """

        try:
            if request_id == 'null':
                request_id = None
            if workload_id == 'null':
                workload_id = None
            if with_request and with_request.lower() in ['true']:
                with_request = True
            else:
                with_request = False
            if with_transform and with_transform.lower() in ['true']:
                with_transform = True
            else:
                with_transform = False
            if with_processing and with_processing.lower() in ['true']:
                with_processing = True
            else:
                with_processing = False

            rets = self.get_requests(request_id=request_id, workload_id=workload_id,
                                     with_request=with_request,
                                     with_transform=with_transform,
                                     with_processing=with_processing)
        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data=rets)

    def post_test(self):
        import pprint
        pprint.pprint(self.get_request())
        pprint.pprint(self.get_request().endpoint)
        pprint.pprint(self.get_request().url_rule)


class MonitorRequest(Monitor):
    """ Monitor Request """

    def get(self, request_id, workload_id):
        """ Get details about a specific Request with given id.
        HTTP Success:
            200 OK
        HTTP Error:
            404 Not Found
            500 InternalError
        :returns: dictionary of an request.
        """

        try:
            if request_id == 'null':
                request_id = None
            if workload_id == 'null':
                workload_id = None

            rets = self.get_requests(request_id=request_id, workload_id=workload_id,
                                     with_request=True, with_transform=False,
                                     with_processing=False)
            status_dict = {'Total': {}}
            min_time, max_time = None, None
            for ret in rets:
                if ret['status'] not in status_dict:
                    status_dict[ret['status']] = {}
                if min_time is None or ret['updated_at'] < min_time:
                    min_time = ret['updated_at']
                if max_time is None or ret['updated_at'] > max_time:
                    max_time = ret['updated_at']

            month_list = self.get_month_list(min_time, max_time)
            for key in status_dict:
                for m in month_list:
                    status_dict[key][m] = 0

            for ret in rets:
                m_time = ret['updated_at'].strftime(r"%Y-%m")
                status_dict['Total'][m_time] += 1
                status_dict[ret['status']][m_time] += 1

            status_dict_acc = {}
            for key in status_dict:
                status_dict_acc[key] = {}
                for i in range(len(month_list)):
                    if i == 0:
                        status_dict_acc[key][month_list[i]] = status_dict[key][month_list[i]]
                    else:
                        status_dict_acc[key][month_list[i]] = status_dict[key][month_list[i]] + status_dict_acc[key][month_list[i - 1]]
            ret_status = {'total': len(rets), 'month_status': status_dict, 'month_acc_status': status_dict_acc}
        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data=ret_status)


class MonitorTransform(Monitor):
    """ Monitor Transform """

    def get(self, request_id, workload_id):
        """ Get details about a specific Request with given id.
        HTTP Success:
            200 OK
        HTTP Error:
            404 Not Found
            500 InternalError
        :returns: dictionary of an request.
        """

        try:
            if request_id == 'null':
                request_id = None
            if workload_id == 'null':
                workload_id = None

            rets = self.get_requests(request_id=request_id, workload_id=workload_id,
                                     with_request=False, with_transform=True, with_processing=False)
            status_dict = {'Total': {}}
            status_dict_by_type = {}
            processed_files, processed_bytes = {}, {}
            processed_files_by_type, processed_bytes_by_type = {}, {}
            min_time, max_time = None, None
            total_files, total_bytes = 0, 0
            for ret in rets:
                if ret['transform_status'] and ret['transform_status'] not in status_dict:
                    status_dict[ret['transform_status']] = {}
                if ret['transform_type'] and ret['transform_type'] not in status_dict_by_type:
                    status_dict_by_type[ret['transform_type']] = {}
                    processed_files_by_type[ret['transform_type']] = {}
                    processed_bytes_by_type[ret['transform_type']] = {}
                if ret['transform_updated_at'] and (min_time is None or ret['transform_updated_at'] < min_time):
                    min_time = ret['transform_updated_at']
                if ret['transform_updated_at'] and (max_time is None or ret['transform_updated_at'] > max_time):
                    max_time = ret['transform_updated_at']

            month_list = self.get_month_list(min_time, max_time)
            for key in status_dict:
                processed_files[key] = {}
                processed_bytes[key] = {}
                for t_type in status_dict_by_type:
                    status_dict_by_type[t_type][key] = {}
                    processed_files_by_type[t_type][key] = {}
                    processed_bytes_by_type[t_type][key] = {}
                for m in month_list:
                    status_dict[key][m] = 0
                    processed_files[key][m] = 0
                    processed_bytes[key][m] = 0
                    for t_type in status_dict_by_type:
                        status_dict_by_type[t_type][key][m] = 0
                        processed_files_by_type[t_type][key][m] = 0
                        processed_bytes_by_type[t_type][key][m] = 0

            for ret in rets:
                if not ret['transform_updated_at']:
                    continue
                m_time = ret['transform_updated_at'].strftime(r"%Y-%m")
                status_dict['Total'][m_time] += 1
                status_dict[ret['transform_status']][m_time] += 1
                processed_files[ret['transform_status']][m_time] += ret['output_processed_files']
                # processed_bytes[ret['transform_status']][m_time] += ret['output_coll_bytes']
                # output_coll_bytes is not filled, need to be fixed on the server
                processed_bytes[ret['transform_status']][m_time] += ret['input_coll_bytes']
                processed_files['Total'][m_time] += ret['output_processed_files']
                # processed_bytes['Total'][m_time] += ret['output_coll_bytes']
                processed_bytes['Total'][m_time] += ret['input_coll_bytes']

                total_files += ret['output_processed_files']
                total_bytes += ret['output_coll_bytes']
                total_bytes += ret['input_coll_bytes']

                t_type = ret['transform_type']
                status_dict_by_type[t_type][ret['transform_status']][m_time] += 1
                processed_files_by_type[t_type][ret['transform_status']][m_time] += ret['output_processed_files']
                # processed_bytes_by_type[t_type][ret['transform_status']][m_time] += ret['output_coll_bytes']
                processed_bytes_by_type[t_type][ret['transform_status']][m_time] += ret['input_coll_bytes']
                status_dict_by_type[t_type]['Total'][m_time] += 1
                processed_files_by_type[t_type]['Total'][m_time] += ret['output_processed_files']
                # processed_bytes_by_type[t_type]['Total'][m_time] += ret['output_coll_bytes']
                processed_bytes_by_type[t_type]['Total'][m_time] += ret['input_coll_bytes']

            status_dict_acc = {}
            processed_files_acc, processed_bytes_acc = {}, {}
            status_dict_by_type_acc = {}
            processed_files_by_type_acc = {}
            processed_bytes_by_type_acc = {}
            for t_type in status_dict_by_type:
                status_dict_by_type_acc[t_type] = {}
                processed_files_by_type_acc[t_type] = {}
                processed_bytes_by_type_acc[t_type] = {}
            for key in status_dict:
                status_dict_acc[key] = {}
                processed_files_acc[key] = {}
                processed_bytes_acc[key] = {}
                for t_type in status_dict_by_type:
                    status_dict_by_type_acc[t_type][key] = {}
                    processed_files_by_type_acc[t_type][key] = {}
                    processed_bytes_by_type_acc[t_type][key] = {}

                for i in range(len(month_list)):
                    if i == 0:
                        status_dict_acc[key][month_list[i]] = status_dict[key][month_list[i]]
                        processed_files_acc[key][month_list[i]] = processed_files[key][month_list[i]]
                        processed_bytes_acc[key][month_list[i]] = processed_bytes[key][month_list[i]]
                        for t_type in status_dict_by_type_acc:
                            status_dict_by_type_acc[t_type][key][month_list[i]] = status_dict_by_type[t_type][key][month_list[i]]
                            processed_files_by_type_acc[t_type][key][month_list[i]] = processed_files_by_type[t_type][key][month_list[i]]
                            processed_bytes_by_type_acc[t_type][key][month_list[i]] = processed_bytes_by_type[t_type][key][month_list[i]]
                    else:
                        status_dict_acc[key][month_list[i]] = status_dict[key][month_list[i]] + status_dict_acc[key][month_list[i - 1]]
                        processed_files_acc[key][month_list[i]] = processed_files[key][month_list[i]] + processed_files_acc[key][month_list[i - 1]]
                        processed_bytes_acc[key][month_list[i]] = processed_bytes[key][month_list[i]] + processed_bytes_acc[key][month_list[i - 1]]
                        for t_type in status_dict_by_type_acc:
                            status_dict_by_type_acc[t_type][key][month_list[i]] = status_dict_by_type[t_type][key][month_list[i]] + status_dict_by_type_acc[t_type][key][month_list[i - 1]]
                            processed_files_by_type_acc[t_type][key][month_list[i]] = processed_files_by_type[t_type][key][month_list[i]] + processed_files_by_type_acc[t_type][key][month_list[i - 1]]
                            processed_bytes_by_type_acc[t_type][key][month_list[i]] = processed_bytes_by_type[t_type][key][month_list[i]] + processed_bytes_by_type_acc[t_type][key][month_list[i - 1]]
            ret_status = {'total': len(rets),
                          'total_files': total_files,
                          'total_bytes': total_bytes,
                          'month_status': status_dict,
                          'month_acc_status': status_dict_acc,
                          'month_processed_files': processed_files,
                          'month_acc_processed_files': processed_files_acc,
                          'month_processed_bytes': processed_bytes,
                          'month_acc_processed_bytes': processed_bytes_acc,
                          'month_status_dict_by_type': status_dict_by_type,
                          'month_acc_status_dict_by_type': status_dict_by_type_acc,
                          'month_processed_files_by_type': processed_files_by_type,
                          'month_acc_processed_files_by_type': processed_files_by_type_acc,
                          'month_processed_bytes_by_type': processed_bytes_by_type,
                          'month_acc_processed_bytes_by_type': processed_bytes_by_type_acc
                          }
        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data=ret_status)


class MonitorProcessing(Monitor):
    """ Monitor Processing """

    def get(self, request_id, workload_id):
        """ Get details about a specific Request with given id.
        HTTP Success:
            200 OK
        HTTP Error:
            404 Not Found
            500 InternalError
        :returns: dictionary of an request.
        """

        try:
            if request_id == 'null':
                request_id = None
            if workload_id == 'null':
                workload_id = None

            rets = self.get_requests(request_id=request_id, workload_id=workload_id,
                                     with_request=False, with_transform=False, with_processing=True)
            status_dict = {'Total': {}}
            min_time, max_time = None, None
            for ret in rets:
                if ret['processing_status'] and ret['processing_status'] not in status_dict:
                    status_dict[ret['processing_status']] = {}
                if ret['processing_updated_at'] and (min_time is None or ret['processing_updated_at'] < min_time):
                    min_time = ret['processing_updated_at']
                if ret['processing_updated_at'] and (max_time is None or ret['processing_updated_at'] > max_time):
                    max_time = ret['processing_updated_at']

            month_list = self.get_month_list(min_time, max_time)
            for key in status_dict:
                for m in month_list:
                    status_dict[key][m] = 0

            for ret in rets:
                if ret['processing_updated_at']:
                    m_time = ret['processing_updated_at'].strftime(r"%Y-%m")
                    status_dict['Total'][m_time] += 1
                    status_dict[ret['processing_status']][m_time] += 1

            status_dict_acc = {}
            for key in status_dict:
                status_dict_acc[key] = {}
                for i in range(len(month_list)):
                    if i == 0:
                        status_dict_acc[key][month_list[i]] = status_dict[key][month_list[i]]
                    else:
                        status_dict_acc[key][month_list[i]] = status_dict[key][month_list[i]] + status_dict_acc[key][month_list[i - 1]]
            ret_status = {'total': len(rets), 'month_status': status_dict, 'month_acc_status': status_dict_acc}
        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data=ret_status)


class MonitorRequestRelation(Monitor):
    """ Monitor Request """

    def get(self, request_id, workload_id):
        """ Get details about a specific Request with given id.
        HTTP Success:
            200 OK
        HTTP Error:
            404 Not Found
            500 InternalError
        :returns: dictionary of an request.
        """

        try:
            if request_id == 'null':
                request_id = None
            if workload_id == 'null':
                workload_id = None

            reqs = get_requests(request_id=request_id, workload_id=workload_id,
                                with_request=True, with_transform=False, with_processing=False,
                                with_detail=False, with_metadata=True)

            for req in reqs:
                req['relation_map'] = []
                workflow = req['request_metadata']['workflow']
                if hasattr(workflow, 'get_relation_map'):
                    req['relation_map'] = workflow.get_relation_map()
            # return reqs
        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        return self.generate_http_response(HTTP_STATUS_CODE.OK, data=reqs)


"""----------------------
   Web service url maps
----------------------"""


def get_blueprint():
    bp = Blueprint('monitor', __name__)

    monitor_view = Monitor.as_view('monitor')
    bp.add_url_rule('/monitor/<request_id>/<workload_id>/<with_request>/<with_transform>/<with_processing>', view_func=monitor_view, methods=['get', ])

    monitor_request_view = MonitorRequest.as_view('monitor_request')
    bp.add_url_rule('/monitor_request/<request_id>/<workload_id>', view_func=monitor_request_view, methods=['get', ])

    monitor_transform_view = MonitorTransform.as_view('monitor_transform')
    bp.add_url_rule('/monitor_transform/<request_id>/<workload_id>', view_func=monitor_transform_view, methods=['get', ])

    monitor_processing_view = MonitorProcessing.as_view('monitor_processing')
    bp.add_url_rule('/monitor_processing/<request_id>/<workload_id>', view_func=monitor_processing_view, methods=['get', ])

    monitor_relation_view = MonitorRequestRelation.as_view('monitor_request_relation')
    bp.add_url_rule('/monitor_request_relation/<request_id>/<workload_id>', view_func=monitor_relation_view, methods=['get', ])

    return bp
